package jaxx.runtime;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Desktop;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Rectangle;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.io.IOException;
import java.io.InputStream;
import java.util.List;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JComboBox;
import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.table.TableCellEditor;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumn;
import javax.swing.text.AbstractDocument;
import javax.swing.text.JTextComponent;
import java.lang.reflect.Field;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.NoSuchElementException;
import java.util.Properties;
import javax.swing.DefaultListCellRenderer;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JLayeredPane;
import javax.swing.JList;
import javax.swing.JRootPane;
import javax.swing.JTabbedPane;
import javax.swing.UIManager;
import javax.swing.UnsupportedLookAndFeelException;
import javax.swing.event.HyperlinkEvent;
import javax.swing.table.DefaultTableCellRenderer;
import jaxx.runtime.swing.BooleanCellRenderer;
import jaxx.runtime.swing.DecoratorTableCellRenderer;
import jaxx.runtime.swing.EmptyNumberTableCellRenderer;
import jaxx.runtime.swing.EnumTableCellRenderer;
import jaxx.runtime.swing.I18nTableCellRenderer;
import jaxx.runtime.swing.Item;
import jaxx.runtime.swing.JAXXComboBox;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jdesktop.jxlayer.JXLayer;

/**
 * The runtime swing util class with some nice stuff.
 *
 * Note : Replace previous class jaxx.runtime.swing.Utils in previous versions.
 * 
 * @author tony
 * @since 1.2
 */
public class SwingUtil extends Util {

    /** to use log facility, just put in your code: log.info(\"...\"); */
    static private final Log log = LogFactory.getLog(SwingUtil.class);
    private static Field numReaders;
    private static Field notifyingListeners;
    public static final String ICON_PREFIX = "icon.";
    public static final String COLOR_PREFIX = "color.";

    public static Dimension newMinDimension() {
        return new Dimension(0, 0);
    }

    public static Dimension newMaxXDimension() {
        return new Dimension(Short.MAX_VALUE, 0);
    }

    public static Dimension newMaxYDimension() {
        return new Dimension(0, Short.MAX_VALUE);
    }

    public static Dimension newMaxXYDimension() {
        return new Dimension(Short.MAX_VALUE, Short.MAX_VALUE);
    }

    public static void setText(final JTextComponent c, final String text) {
        try {
            // AbstractDocument deadlocks if we try to acquire a write lock while a read lock is held by the current thread
            // If there are any readers, dispatch an invokeLater.  This should only happen in the event of circular bindings.
            // Similarly, circular bindings can result in an "Attempt to mutate in notification" error, which we deal with
            // by checking for the 'notifyingListeners' property.
            AbstractDocument document = (AbstractDocument) c.getDocument();
            if (numReaders == null) {
                numReaders = AbstractDocument.class.getDeclaredField("numReaders");
                numReaders.setAccessible(true);
            }
            if (notifyingListeners == null) {
                notifyingListeners = AbstractDocument.class.getDeclaredField("notifyingListeners");
                notifyingListeners.setAccessible(true);
            }

            if (notifyingListeners.get(document).equals(Boolean.TRUE)) {
                return;
            }

            if ((Integer) numReaders.get(document) > 0) {
                SwingUtilities.invokeLater(new Runnable() {

                    @Override
                    public void run() {
                        if (!c.getText().equals(text)) {
                            c.setText(text);
                        }
                    }
                });
                return;
            }

            String oldText = c.getText();
            if (oldText == null || !oldText.equals(text)) {
                c.setText(text);
            }
        } catch (NoSuchFieldException e) {
            throw new RuntimeException(e);
        } catch (IllegalAccessException e) {
            throw new RuntimeException(e);
        } catch (SecurityException e) {
            c.setText(text);
        }
    }

    /**
     * Fill a combo box model with some datas, and select after all the given object
     *
     * @param combo  the combo to fill
     * @param data   data ot inject in combo
     * @param select the object to select in combo after reflling his model
     */
    public static void fillComboBox(JComboBox combo, Collection<?> data, Object select) {
        if (!(combo.getModel() instanceof DefaultComboBoxModel)) {
            throw new IllegalArgumentException("this method need a DefaultComboBoxModel for this model but was " + combo.getModel().getClass());
        }
        DefaultComboBoxModel model = (DefaultComboBoxModel) combo.getModel();
        // evince the model
        model.removeListDataListener(combo);
        model.removeAllElements();
        if (data != null) {
            for (Object o : data) {
                model.addElement(o);
            }
        }
        // attach the model
        model.addListDataListener(combo);
        model.setSelectedItem(select);
    }

    /**
     * Fill a combo box model with some datas, and select after all the given object
     *
     * @param combo     the combo to fill
     * @param data      data ot inject in combo
     * @param select    the object to select in combo after reflling his model
     * @param firstNull add a first null element
     */
    public static void fillComboBox(JAXXComboBox combo, Collection<?> data, Object select, boolean firstNull) {
        List<Item> items = new ArrayList<Item>();
        if (firstNull) {
            items.add(new Item("null", " ", null, false));
        }
        if (data != null) {
            for (Object d : data) {
                items.add(new Item(d.toString(), d.toString(), d, d.equals(select)));
            }
        }
        combo.setItems(items);
    }

    public static int computeTableColumnWidth(JTable table, Font font, int columnIndex, String suffix) {
        int width = 0;
        if (font == null) {
            font = table.getFont();
        }
//        if (font == null) {
//            TableColumn column = table.getColumnModel().getColumn(columnIndex);
//            font = ((JComponent) column.getCellRenderer()).getFont();
//        }
        FontMetrics fontMetrics = table.getFontMetrics(font);
        for (int i = 0, rowCount = table.getRowCount(); i < rowCount; i++) {
            String key = (String) table.getModel().getValueAt(i, 0);
            int w = fontMetrics.stringWidth(key + suffix);
            if (w > width) {
                width = w;
            }
        }
        return width;
    }

    public static void fixTableColumnWidth(JTable table, int columnIndex, int width) {
        TableColumn column = table.getColumnModel().getColumn(columnIndex);
        column.setMaxWidth(width);
        column.setMinWidth(width);
        column.setWidth(width);
        column.setPreferredWidth(width);
    }

    public static void setTableColumnEditor(JTable table, int columnIndex, TableCellEditor editor) {
        TableColumn column = table.getColumnModel().getColumn(columnIndex);
        column.setCellEditor(editor);
    }

    public static void setTableColumnRenderer(JTable table, int columnIndex, TableCellRenderer editor) {
        TableColumn column = table.getColumnModel().getColumn(columnIndex);
        column.setCellRenderer(editor);
    }

    public static void setI18nTableHeaderRenderer(JTable table, String... libelles) {
        table.getTableHeader().setDefaultRenderer(new I18nTableCellRenderer(table.getTableHeader().getDefaultRenderer(), libelles));
    }

    public static TableCellRenderer newStringTableCellRenderer(final DefaultTableCellRenderer renderer, final int length, final boolean tooltip) {

        return new DefaultTableCellRenderer() {

            private static final long serialVersionUID = 1l;

            @Override
            public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column) {

                renderer.getTableCellRendererComponent(table, value, isSelected, hasFocus, row, column);
                String val = renderer.getText();
                String val2 = val;
                if (val.length() > length) {
                    val2 = val.substring(0, length - 3) + "...";
                }

                JComponent comp = (JComponent) super.getTableCellRendererComponent(table, val2, isSelected, hasFocus, row, column);
                if (tooltip) {
                    comp.setToolTipText(val);
                }
                return comp;
            }
        };
    }

    /**
     * Box a component in a {@link org.jdesktop.jxlayer.JXLayer}.
     *
     * @param component the component to box
     * @return the {@link org.jdesktop.jxlayer.JXLayer} boxing the component
     */
    public static JXLayer<?> boxComponentWithJxLayer(JComponent component) {
        JXLayer<?> layer = getLayer(component);
        if (layer != null) {
            return layer;
        }
        layer = new org.jdesktop.jxlayer.JXLayer<JComponent>();
        layer.add(component);
        return layer;
    }

    public static List<JComponent> getLayeredComponents(JAXXObject object) {
        List<JComponent> result = new ArrayList<JComponent>();
        for (Entry<String, Object> child : object.get$objectMap().entrySet()) {
            if (child.getValue() == null) {
                log.warn("find a null object in $objectMap " + child.getKey());
                continue;
            }
            if (JComponent.class.isAssignableFrom(child.getValue().getClass())) {
                JComponent comp = (JComponent) child.getValue();
                if (isLayered(comp)) {
                    result.add(comp);
                }
            }
        }
        return result;
    }

    @SuppressWarnings("unchecked")
    public static JXLayer<JComponent> getLayer(JComponent comp) {
        if (!isLayered(comp)) {
            return null;
        }
        return (JXLayer<JComponent>) comp.getParent();
    }

    public static boolean isLayered(JComponent comp) {
        Container parent = comp.getParent();
        return parent != null && parent instanceof JXLayer<?>;
    }

    /**
     * recherche les composants portant le meme nom que les champs de la classe
     * clazz. Cette methode est statique pour pouvoir eventuellement l'utiliser
     * dans un autre context (je pense par exemple a la generation jaxx).
     * <p/>
     * <p/>
     * Si la recherche echoue pour quelque raison que se soit, aucune exception
     * n'est leve, et la map retournee est tout simplement vide ou incomplete
     *
     * @param clazz     la classe ou recherche les champs
     * @param container le container ou rechercher les composants d'edition
     * @return le dictionnaire des composants recherches.
     */
    public static Map<String, JComponent> lookingForEditor(Class<?> clazz, Container container) {
        Map<String, JComponent> result = new HashMap<String, JComponent>();
        try {
            // looking for all component with name set
            Map<String, JComponent> allNamedComponent = new HashMap<String, JComponent>();
            List<Container> todo = new LinkedList<Container>();
            todo.add(container);
            while (todo.size() > 0) {
                for (ListIterator<Container> i = todo.listIterator(); i.hasNext();) {
                    Container parent = i.next();
                    i.remove();
                    for (Component c : parent.getComponents()) {
                        if (c instanceof Container) {
                            i.add((Container) c);
                            String name = c.getName();
                            if (c instanceof JComponent &&
                                    name != null && !"".equals(name)) {
                                allNamedComponent.put(name, (JComponent) c);
                            }
                        }
                    }
                }
            }

            // looking for all properties on class
            BeanInfo info = Introspector.getBeanInfo(clazz);
            PropertyDescriptor[] props = info.getPropertyDescriptors();

            // find if one properties have same name that component
            for (PropertyDescriptor prop : props) {
                String name = prop.getName();
                if (allNamedComponent.containsKey(name)) {
                    result.put(name, allNamedComponent.get(name));
                }
            }

        } catch (IntrospectionException eee) {
            log.warn("Can't introspect bean", eee);
        }

        if (log.isDebugEnabled()) {
            log.debug("Result: " + result);
        }

        return result;
    }

    /**
     * Centrer un component graphique au center d'un autre component.
     *
     * <b>Note:</b> si le parent est null, alors on ne fait rien.
     *
     * @param parent le component parent
     * @param component le component à centrer
     */
    public static void center(Component parent, Component component) {
        if (parent == null) {
            return;
        }
        Rectangle r = parent.getBounds();
        int x = r.x + (r.width - component.getSize().width) / 2;
        int y = r.y + (r.height - component.getSize().height) / 2;
        component.setLocation(x, y);
    }

    /**
     * Try to load the Nimbus look and feel.
     * <p/>
     * @throws UnsupportedLookAndFeelException
     *                                if nimbus is not applicable
     * @throws ClassNotFoundException
     * @throws InstantiationException
     * @throws IllegalAccessException
     */
    public static void initNimbusLoookAndFeel() throws UnsupportedLookAndFeelException, ClassNotFoundException, InstantiationException, IllegalAccessException {

        for (UIManager.LookAndFeelInfo laf : UIManager.getInstalledLookAndFeels()) {
            if ("Nimbus".equals(laf.getName())) {
                UIManager.setLookAndFeel(laf.getClassName());
            }
        }
    }

    /**
     * Load the ui.properties file and set in {@link UIManager} colors and icons found.
     *
     * @param defaultUIConfig le path vers le fichier de la config d'ui par défaut (doit etre dansle class-path)
     * @param extraUIConfig le path vers une surcharge de la config d'ui (doit etre dans le class-path)
     * 
     * @throws IOException            if could not load the ui.properties file
     */
    public static void loadUIConfig(String defaultUIConfig, String extraUIConfig) throws IOException {

        Properties p = new Properties();
        log.info("loading default UI config " + defaultUIConfig);
        p.load(SwingUtil.class.getResourceAsStream(defaultUIConfig));
        if (log.isDebugEnabled()) {
            log.debug(p.toString());
        }
        if (extraUIConfig != null) {
            InputStream extraStream = SwingUtil.class.getResourceAsStream(extraUIConfig);
            if (extraStream == null) {
                log.warn("could not find extraUIConfig : " + extraUIConfig);
            } else {
                log.info("loading extra UI config " + extraUIConfig);
                Properties p2 = new Properties(p);
                p2.load(extraStream);
                if (log.isDebugEnabled()) {
                    log.debug(p2.toString());
                }
                p.putAll(p2);
            }
        }
        for (Entry<Object, Object> entry : p.entrySet()) {
            String key = (String) entry.getKey();
            if (key.startsWith(ICON_PREFIX)) {
                ImageIcon icon;
                try {
                    icon = createImageIcon((String) entry.getValue());
                    UIManager.put(key.substring(ICON_PREFIX.length()), icon);
                } catch (Exception e) {
                    log.warn("could not load icon " + entry.getValue());
                }
                continue;
            }
            if (key.startsWith(COLOR_PREFIX)) {
                String value = (String) entry.getValue();
                String[] rgb = value.split(",");
                UIManager.put(key.substring(COLOR_PREFIX.length()), new Color(Integer.valueOf(rgb[0]), Integer.valueOf(rgb[1]), Integer.valueOf(rgb[2])));
            }
        }
    }

    /**
     * Iterate the components of a {@link JTabbedPane} in natural order.
     *
     * Says using method {@link JTabbedPane#getComponent(int)}
     * @param tabs the
     * @return
     * @since 1.4
     */
    public static TabbedPaneIterator<Component> newTabbedPaneIterator(JTabbedPane tabs) {
        return new TabbedPaneIterator<Component>(false, tabs) {

            @Override
            protected Component get(int index, Component comp) {
                return comp;
            }
        };
    }

    /**
     * A simple iterator on a {@link JTabbedPane}.
     *
     * Implements the method {@link #get(int, java.awt.Component)} to obtain
     * the data required given the component (or index).
     *
     * You can also inverse the order by usin the method {@link #reverse()}.
     *
     * Note: After the use of the method {@link #reverse()} the iterator returns
     * to the first element.
     * 
     * @param <O> the type of return elements.
     * @since 1.4
     */
    public static abstract class TabbedPaneIterator<O> implements Iterator<O> {

        final JTabbedPane tabs;
        boolean reverse;
        int index;
        int increment;

        protected abstract O get(int index, Component comp);

        public TabbedPaneIterator(boolean reverse, JTabbedPane tabs) {
            this.tabs = tabs;
            setReverse(reverse);
        }

        public void reset() {
            setReverse(reverse);
        }

        public int size() {
            return tabs.getTabCount();
        }

        public TabbedPaneIterator<O> reverse() {
            setReverse(!reverse);
            return this;
        }

        @Override
        public boolean hasNext() {
            return reverse ? index > 0 : index < tabs.getTabCount();
        }

        public int getIndex() {
            return index;
        }

        @Override
        public O next() {
            if (!hasNext()) {
                throw new NoSuchElementException();
            }
            Component next = tabs.getComponentAt(index);
            O result = get(index, next);
            index += increment;
            return result;
        }

        @Override
        public void remove() {
            throw new IllegalStateException("not implemented for " + this);
        }

        @Override
        public String toString() {
            return super.toString() + "< reverse:" + reverse + ", index:" + index + ", size:" + tabs.getTabCount() + " >";
        }

        protected void setReverse(boolean reverse) {
            if (reverse) {
                index = tabs.getTabCount() - 1;
                increment = -1;
            } else {
                index = 0;
                increment = 1;
            }
            this.reverse = reverse;
        }
    }

    public static JLabel newLabel(String text, Object iconKey, int aligment) {
        Icon icon = null;
        if (iconKey instanceof Icon) {
            icon = (Icon) iconKey;
        } else if (iconKey instanceof String) {
            icon = jaxx.runtime.Util.getUIManagerActionIcon((String) iconKey);
        }
        JLabel result;
        if (icon == null) {
            result = new JLabel(text, aligment);
        } else {
            result = new JLabel(text, icon, aligment);
        }
        return result;
    }

    /**
     * Gets the higest visible component in a ancestor hierarchy at
     * specific x,y coordinates
     * @param parent
     * @param x
     * @param y
     * @return
     */
    public static Component getDeepestObjectAt(Component parent, int x, int y) {

        if (parent instanceof Container) {
            Container cont = (Container) parent;
            // use a copy of 1.3 Container.findComponentAt
            Component child = findComponentAt(cont, cont.getWidth(), cont.getHeight(), x, y);
            if (child != null && child != cont) {
                //log.info("child find : " + child.getName());
                if (child instanceof JRootPane) {
                    JLayeredPane lp = ((JRootPane) child).getLayeredPane();
                    Rectangle b = lp.getBounds();
                    child = getDeepestObjectAt(lp, x - b.x, y - b.y);
                }
                if (child != null) {
                    return child;
                }
            }
        }
        // if the parent is not a Container then it might be a MenuItem.
        // But even if it isn't a MenuItem just return the parent because
        // that's a close as we can come.
        return parent;
    }

    public static Component findComponentAt(Container cont, int width, int height, int x, int y) {
        //log.info("container : " + cont.getName());
        synchronized (cont.getTreeLock()) {

            if (!((x >= 0) && (x < width) && (y >= 0) && (y < height) && cont.isVisible() && cont.isEnabled())) {
                return null;
            }

            Component[] component = cont.getComponents();
            int ncomponents = cont.getComponentCount();

            // Two passes: see comment in sun.awt.SunGraphicsCallback
            for (int i = 0; i < ncomponents; i++) {
                Component comp = component[i];
                Rectangle rect = null;

                if (comp != null && !comp.isLightweight()) {
                    if (rect == null || rect.width == 0 || rect.height == 0) {
                        rect = comp.getBounds();
                    }
                    if (comp instanceof JXLayer<?>) {
                        JXLayer<?> layer = (JXLayer<?>) comp;
                        comp = layer.getView();
                    }
                    if (comp instanceof Container) {
                        comp = findComponentAt((Container) comp, rect.width, rect.height, x - rect.x, y - rect.y);
                    } else {
                        comp = comp.getComponentAt(x - rect.x, y - rect.y);
                    }
                    if (comp != null && comp.isVisible() && comp.isEnabled()) {
                        return comp;
                    }
                }
            }

            for (int i = 0; i < ncomponents; i++) {
                Component comp = component[i];
                Rectangle rect = null;

                if (comp != null && comp.isLightweight()) {
                    if (rect == null || rect.width == 0 || rect.height == 0) {
                        rect = comp.getBounds();
                    }
                    if (comp instanceof JXLayer<?>) {
                        JXLayer<?> layer = (JXLayer<?>) comp;
                        comp = layer.getView();
                    }
                    if (comp instanceof Container) {
                        comp = findComponentAt((Container) comp, rect.width, rect.height, x - rect.x, y - rect.y);
                    } else {
                        comp = comp.getComponentAt(x - rect.x, y - rect.y);
                    }
                    if (comp != null && comp.isVisible() && comp.isEnabled()) {
                        return comp;
                    }
                }
            }
            return cont;
        }
    }

    /**
     * Génère un renderer de liste graphique basée sur un décorateur donné.
     *
     * @param <O> le type des données gérées par le décorateur
     * @param decorator le décorateur à encapsuler
     * @return le nouveau renderer
     * @see Decorator
     */
    public static <O> DefaultListCellRenderer newDecoratedListCellRenderer(final Decorator<O> decorator) {
        return new DefaultListCellRenderer() {

            private static final long serialVersionUID = 1L;

            @Override
            public Component getListCellRendererComponent(JList list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
                String decorated;
                if (value instanceof String) {
                    decorated = (String) value;
                } else {
                    decorated = decorator.toString(value);
                }
                JComponent component = (JComponent) super.getListCellRendererComponent(list, decorated, index, isSelected, cellHasFocus);
                return component;
            }
        };
    }

    public static TableCellRenderer newDeleteCellRenderer(DefaultTableCellRenderer renderer) {
        Icon icon = UIManager.getIcon("Table.removeIcon");
        if (icon == null) {
            // try with default icon
            icon = createActionIcon("delete");
        }
        return new BooleanCellRenderer(renderer, icon);
    }

    public static TableCellRenderer newBooleanTableCellRenderer(DefaultTableCellRenderer renderer) {
        return new BooleanCellRenderer(renderer);
    }

    public static DecoratorTableCellRenderer newDecorateTableCellRenderer(TableCellRenderer renderer, Decorator<?> decorator) {
        return new DecoratorTableCellRenderer(renderer, decorator);
    }

    public static EmptyNumberTableCellRenderer newEmptyNumberTableCellRenderer(TableCellRenderer renderer) {
        return new EmptyNumberTableCellRenderer(renderer);
    }

    public static <E extends Enum<E>> EnumTableCellRenderer<E> newEnumTableCellRenderer(TableCellRenderer renderer, Class<E> enumClass) {
        return new EnumTableCellRenderer<E>(renderer, enumClass);
    }

    /**
     * Open a link coming from a {@link HyperlinkEvent}.
     *
     * And try to open the link if an url in a browser.
     * 
     * @param he the event to treate
     * @since 1.6.0
     */
    public static void openLink(HyperlinkEvent he) {
        if (he.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {

            if (Desktop.isDesktopSupported()) {
                try {
                    URL u = he.getURL();
                    if (u.getProtocol().equalsIgnoreCase("mailto") || u.getProtocol().equalsIgnoreCase("http") || u.getProtocol().equalsIgnoreCase("ftp")) {
                        Desktop.getDesktop().browse(u.toURI());
                    }
                } catch (IOException e) {
                    if (log.isErrorEnabled()) {
                        log.error("Error while opening link", e);
                    }
                } catch (URISyntaxException e) {
                    if (log.isErrorEnabled()) {
                        log.error("Error while opening link", e);
                    }
                }
            }
        }
    }
}
