package jaxx.runtime.validator.swing;

import jaxx.runtime.validator.BeanValidatorEvent;

import javax.swing.JComponent;
import java.util.ArrayList;
import java.util.List;
import jaxx.runtime.validator.BeanValidatorField;
import jaxx.runtime.validator.BeanValidatorScope;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * The model of the list of validation's messages
 *
 * @author chemit
 */
public class SwingValidatorMessageListModel
        extends javax.swing.AbstractListModel
        implements jaxx.runtime.validator.BeanValidatorListener {

    private static final long serialVersionUID = 1L;
    /** to use log facility, just put in your code: log.info(\"...\"); */
    private static Log log = LogFactory.getLog(SwingValidatorMessageListModel.class);
    /** list of registred validators */
    protected transient List<SwingValidator<?>> validators;
    /** list of messages actual displayed */
    protected List<SwingValidatorMessage> data;

    public SwingValidatorMessageListModel() {
        validators = new ArrayList<SwingValidator<?>>();
        data = new java.util.ArrayList<SwingValidatorMessage>();
    }

    public boolean isEmpty() {
        return getSize() == 0;
    }

    public void registerValidator(SwingValidator<?> validator) {
        if (validators.contains(validator)) {
            throw new IllegalArgumentException("the validator " + validator + " is already registred in " + this);
        }
        validators.add(validator);
        validator.addBeanValidatorListener(this);
    }

    public void clear() {
        int i = data.size();
        if (i > 0) {
            data.clear();
            fireIntervalRemoved(this, 0, i - 1);
        }
    }

    @Override
    public int getSize() {
        return data.size();
    }

    @Override
    public Object getElementAt(int index) {
        ensureRowIndex(index);
        return data.get(index);
    }

    @Override
    public void onFieldChanged(BeanValidatorEvent event) {
        String[] toDelete = event.getMessagesToDelete();
        String[] toAdd = event.getMessagesToAdd();
        BeanValidatorField field = event.getField();
        BeanValidatorScope scope = event.getScope();
        boolean mustAdd = toAdd != null && toAdd.length > 0;
        boolean mustDel = toDelete != null && toDelete.length > 0;

        if (log.isTraceEnabled()) {
            log.trace("----------------------------------------------------------");
            log.trace(field + " - (" + getSize() + ") toAdd     " + mustAdd);
            log.trace(field + " - (" + getSize() + ") toDelete  " + mustDel);
        }

        SwingValidator validator = (SwingValidator) event.getSource();

        if (mustDel) {

            // removes datas and notify if no messages to add
            removeMessages(validator, field, scope, !mustAdd, toDelete);
        }

        if (mustAdd) {

            // add new messages, sort datas and notify
            addMessages(validator, field, scope, true, toAdd);
        }
    }

    protected void ensureRowIndex(int index) throws ArrayIndexOutOfBoundsException {
        if (index < -1 || index >= getSize()) {
            throw new ArrayIndexOutOfBoundsException("the rowIndex was " + index + ", but should be int [0," + (getSize() - 1) + "]");
        }
    }

    protected void addMessages(SwingValidator validator, BeanValidatorField field, BeanValidatorScope scope, boolean sort, String... messages) {

        JComponent editor = validator.getFieldRepresentation(field.getName());
        // add new errors
        for (String error : messages) {
            SwingValidatorMessage row = new SwingValidatorMessage(validator, field, error, scope, editor);
            data.add(row);
            if (!sort) {
                fireIntervalAdded(this, data.size() - 1, data.size() - 1);
            }
        }

        if (sort) {

            // resort datas
            java.util.Collections.sort(data);

            // notify
            fireContentsChanged(this, 0, getSize() - 1);
        }
    }

    protected void removeMessages(SwingValidator validator, BeanValidatorField field, BeanValidatorScope scope, boolean notify, String... messages) {

        List<String> messagesToDel = new java.util.ArrayList<String>(java.util.Arrays.asList(messages));

        // do it in reverse mode (only one pass in that way since index
        // will stay coherent while removing them)

        for (int i = getSize() - 1; i > -1; i--) {
            SwingValidatorMessage error = data.get(i);
            if (error.getValidator() == validator && error.getScope() == scope && error.getField() == field && messagesToDel.contains(error.getMessage())) {
                // remove the message
                data.remove(i);
                if (notify) {
                    fireIntervalRemoved(this, i, i);
                }
            }
        }
    }
}
